#include <GL/gl.h>
#include <GL/glu.h>
#include <SDL/SDL.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define NUM_HALL_SEGMENTS 10
#define WIDTH 640
#define HEIGHT 480
#define FULLSCREEN
#define NEAR 0.1
#define FAR 120.0
#define XSENS 0.2
#define YSENS 0.2
#define VELOCITY 8.0
#define FILTER_CONST 8.0
#define XCLIP 1.8
#define YCLIP 3.18
#define ZCLIP ((float)NUM_HALL_SEGMENTS*4.3+2.0)
#undef USE_LIGHTING
#define FRAME_TIME 28   // milliseconds, ~=35fps

#define NUM_TEXTURES 4
#define TEX_WOOD    0
#define TEX_WALL    1
#define TEX_CARPET  2
#define TEX_CEILING 3

#define PI_OVER_180 (3.141592653589/180.0)

int num_lights=NUM_HALL_SEGMENTS;

const GLfloat LightAmbient[]= {0.6,0.6,0.6,1.0},
              LightDiffuse[]= {1.0,1.0,1.0,1.0},
              LightPosition[]={0.0,3.0,2.0,1.0};

GLuint textures[NUM_TEXTURES];

struct ViewState
{
    GLfloat xpos,ypos,zpos,
            xang,yang,
            xvel,yvel,zvel;
};

SDL_Surface *LoadBMP(char *filename)
{
    Uint8 *rowhi, *rowlo;
    Uint8 *tmpbuf, tmpch;
    SDL_Surface *image;
    int i, j;

    image = SDL_LoadBMP(filename);
    if ( image == NULL ) {
        fprintf(stderr, "Unable to load %s: %s\n", filename, SDL_GetError());
        return(NULL);
    }

    /* GL surfaces are upsidedown and RGB, not BGR :-) */
    tmpbuf = (Uint8 *)malloc(image->pitch);
    if ( tmpbuf == NULL ) {
        fprintf(stderr, "Out of memory\n");
        return(NULL);
    }
    rowhi = (Uint8 *)image->pixels;
    rowlo = rowhi + (image->h * image->pitch) - image->pitch;
    for ( i=0; i<image->h/2; ++i ) {
        for ( j=0; j<image->w; ++j ) {
            tmpch = rowhi[j*3];
            rowhi[j*3] = rowhi[j*3+2];
            rowhi[j*3+2] = tmpch;
            tmpch = rowlo[j*3];
            rowlo[j*3] = rowlo[j*3+2];
            rowlo[j*3+2] = tmpch;
        }
        memcpy(tmpbuf, rowhi, image->pitch);
        memcpy(rowhi, rowlo, image->pitch);
        memcpy(rowlo, tmpbuf, image->pitch);
        rowhi += image->pitch;
        rowlo -= image->pitch;
    }
    free(tmpbuf);
    return(image);
}

void InitTextures(void)
{
    SDL_Surface *image;
    glGenTextures(3,textures);
    image=LoadBMP("wood.bmp");
    if(!image)
    {
        SDL_Quit();
        exit(1);
    }
    glBindTexture(GL_TEXTURE_2D,textures[TEX_WOOD]);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
      GL_LINEAR_MIPMAP_LINEAR);
    gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGB,image->w,image->h,GL_RGB,
      GL_UNSIGNED_BYTE,image->pixels);
    SDL_FreeSurface(image);
    image=LoadBMP("wall.bmp");
    if(!image)
    {
        SDL_Quit();
        exit(1);
    }
    glBindTexture(GL_TEXTURE_2D,textures[TEX_WALL]);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
      GL_LINEAR_MIPMAP_LINEAR);
    gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGB,image->w,image->h,GL_RGB,
      GL_UNSIGNED_BYTE,image->pixels);
    SDL_FreeSurface(image);
    image=LoadBMP("carpet.bmp");
    if(!image)
    {
        SDL_Quit();
        exit(1);
    }
    glBindTexture(GL_TEXTURE_2D,textures[TEX_CARPET]);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
      GL_LINEAR_MIPMAP_LINEAR);
    gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGB,image->w,image->h,GL_RGB,
      GL_UNSIGNED_BYTE,image->pixels);
    SDL_FreeSurface(image);
    image=LoadBMP("ceiling.bmp");
    if(!image)
    {
        SDL_Quit();
        exit(1);
    }
    glBindTexture(GL_TEXTURE_2D,textures[TEX_CEILING]);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
      GL_LINEAR_MIPMAP_LINEAR);
    gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGB,image->w,image->h,GL_RGB,
      GL_UNSIGNED_BYTE,image->pixels);
    SDL_FreeSurface(image);
}

int InitGL(int width,int height)
{
//    SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE,32);
//    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,24);
//    SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE,0);
//    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
    if(SDL_SetVideoMode(width,height,0,SDL_OPENGL
#ifdef FULLSCREEN
      |SDL_FULLSCREEN
#endif
      )==NULL)
    {
        fprintf(stderr,"Unable to create OpenGL screen: %s\n",SDL_GetError());
        SDL_Quit();
        return 2;
    }
    glViewport(0,0,width,height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0,(GLfloat)WIDTH/(GLfloat)HEIGHT,NEAR,FAR);
    glMatrixMode(GL_MODELVIEW);
    glClearColor(0.0,0.0,0.0,0.0);
    glClearDepth(1.0);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
    glShadeModel(GL_SMOOTH);
    glDepthFunc(GL_LESS);
    glEnable(GL_DEPTH_TEST);

    InitTextures();
    glEnable(GL_TEXTURE_2D);

#ifdef USE_LIGHTING
    glLightfv(GL_LIGHT1,GL_AMBIENT,LightAmbient);
    glLightfv(GL_LIGHT1,GL_DIFFUSE,LightDiffuse);
    glLightfv(GL_LIGHT1,GL_POSITION,LightPosition);
    glEnable(GL_LIGHT1);
    glEnable(GL_LIGHTING);
#endif
    return 0;
}

float scaleRawValue(int a)
{
	return .2+(float)a/319.0;
}

void DrawCeilingSect(int a,int b)
{
	float t;

    glBindTexture(GL_TEXTURE_2D,textures[TEX_CEILING]);
	
	t=scaleRawValue(a);

	glBegin(GL_QUADS);
	glColor3f(t,t,t);
	glTexCoord2f(0.0,0.0);glVertex3f(-1.952,3.3,4.3);
	glTexCoord2f(.5,0);glVertex3f(0,3.3,4.3);
	glTexCoord2f(.5,-1.0);glVertex3f(0,3.3,-4.3);
	glTexCoord2f(0.0,-1.0);glVertex3f(-1.952,3.3,-4.3);
	glEnd();

	t=scaleRawValue(b);

	glBegin(GL_QUADS);
	glColor3f(t,t,t);
	glTexCoord2f(.5,0.0);glVertex3f(0,3.3,4.3);
	glTexCoord2f(1.0,0);glVertex3f(1.952,3.3,4.3);
	glTexCoord2f(1.0,-1.0);glVertex3f(1.952,3.3,-4.3);
	glTexCoord2f(.5,-1.0);glVertex3f(0,3.3,-4.3);
	glEnd();

#if 0
	glBegin(GL_QUADS);
	glColor3f(.2,.2,.2);
	glTexCoord2f(0.0,-.7466);glVertex3f(-1.952,3.3,-2.12);
	glTexCoord2f(1.0,-.7466);glVertex3f(1.952,3.3,-2.12);
	glTexCoord2f(1.0,-1.0);glVertex3f(1.952,3.3,-4.3);
	glTexCoord2f(0,-1.0);glVertex3f(-1.952,3.3,-4.3);
	glEnd();
#endif
}

void DrawWallSect(int a, int b)
{
	float t;

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,textures[TEX_WALL]);

	t=scaleRawValue(b);

	glBegin(GL_QUADS);
	glColor3f(t,t,t);
	glNormal3f(0,0,1.0);

	glTexCoord2f(0.03,2.2);glVertex3f(-3.211,3.301,0.0);
	glTexCoord2f(0.03,1.1);glVertex3f(-3.211,0,0.0);
	glTexCoord2f(1.7,1.1);glVertex3f(3.211,0,0.0);
	glTexCoord2f(1.7,2.2);glVertex3f(3.211,3.301,0.0);
		
	glEnd();
	
	t=scaleRawValue(a);
	
	glBegin(GL_QUADS);
	glColor3f(t,t,t);
	glNormal3f(0,0,1.0);

	glTexCoord2f(0.03,1.1);glVertex3f(-3.211,0,0.0);
	glTexCoord2f(0.03,0);glVertex3f(-3.211,-3.301,0.0);
	glTexCoord2f(1.7,0);glVertex3f(3.211,-3.301,0.0);
	glTexCoord2f(1.7,1.1);glVertex3f(3.211,0,0.0);

	glEnd();

/*    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(1.0,1.0,1.0);
    glNormal3f(0.0,0.0,1.0);
    glTexCoord2f(0.03,2.2);glVertex3f(-3.211,3.301,0.0);
    glTexCoord2f(0.03,0.0);glVertex3f(-3.211,-3.301,0.0);
    glTexCoord2f(1.0,2.2);glVertex3f(-0.3,3.301,0.0);
    glTexCoord2f(1.0,0.0);glVertex3f(-0.3,-3.301,0.0);
    glTexCoord2f(1.3333333,2.2);glVertex3f(0.3,3.301,0.0);
    glTexCoord2f(1.3333333,0.0);glVertex3f(0.3,-3.301,0.0);
    glTexCoord2f(2.3033333,2.2);glVertex3f(3.211,3.301,0.0);
    glTexCoord2f(2.3033333,0.0);glVertex3f(3.211,-3.301,0.0);
    glEnd();
    */
}

void DrawDoorSect(int a)
{
    float t=scaleRawValue(a);
    // Draw wall above doorframe
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,textures[TEX_WALL]);
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(t,t,t);
    glNormal3f(0.0,0.0,1.0);
    glTexCoord2f(2.3033333,0.0);glVertex3f(-1.091,-3.301,0.0);
    glTexCoord2f(2.3033333,2.2);glVertex3f(-1.091,3.301,0.0);
    glTexCoord2f(2.3033333,2.05);glVertex3f(-1.091,2.85,0.0);
    glTexCoord2f(3.03,2.2);glVertex3f(1.091,3.301,0.0);
    glTexCoord2f(3.03,2.05);glVertex3f(1.091,2.85,0.0);
    glTexCoord2f(3.03,0.0);glVertex3f(1.091,-3.301,0.0);
    glEnd();
    
    // Draw left part of doorframe
    glDisable(GL_TEXTURE_2D);
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.34,0.31,0.20);
    glNormal3f(-1.0,0.0,0.0);
    glVertex3f(-1.091,2.85,0.0);
    glVertex3f(-1.091,-3.301,0.0);
    glVertex3f(-1.091,2.85,0.03);
    glVertex3f(-1.091,-3.301,0.03);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.30,0.27,0.16);
    glNormal3f(0.0,0.0,1.0);
    glVertex3f(-1.091,2.85,0.03);
    glVertex3f(-1.091,-3.301,0.03);
    glVertex3f(-1.0,2.85,0.03);
    glVertex3f(-1.0,-3.301,0.03);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.26,0.23,0.12);
    glNormal3f(1.0,0.0,0.0);
    glVertex3f(-1.0,2.85,0.03);
    glVertex3f(-1.0,-3.301,0.03);
    glVertex3f(-1.0,2.85,-0.03);
    glVertex3f(-1.0,-3.301,-0.03);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.24,0.21,0.10);
    glNormal3f(0.0,0.0,1.0);
    glVertex3f(-1.0,2.85,-0.03);
    glVertex3f(-1.0,-3.301,-0.03);
    glVertex3f(-0.97,2.85,-0.03);
    glVertex3f(-0.97,-3.301,-0.03);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.26,0.23,0.12);
    glNormal3f(1.0,0.0,0.0);
    glVertex3f(-0.97,2.85,-0.03);
    glVertex3f(-0.97,-3.301,-0.03);
    glVertex3f(-0.97,2.85,-0.09);
    glVertex3f(-0.97,-3.301,-0.09);
    glEnd();
    
    // Draw right part of doorframe
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.34,0.31,0.20);
    glNormal3f(1.0,0.0,0.0);
    glVertex3f(1.091,2.85,0.0);
    glVertex3f(1.091,-3.301,0.0);
    glVertex3f(1.091,2.85,0.03);
    glVertex3f(1.091,-3.301,0.03);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.30,0.27,0.16);
    glNormal3f(0.0,0.0,1.0);
    glVertex3f(1.091,2.85,0.03);
    glVertex3f(1.091,-3.301,0.03);
    glVertex3f(1.0,2.85,0.03);
    glVertex3f(1.0,-3.301,0.03);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.26,0.23,0.12);
    glNormal3f(-1.0,0.0,0.0);
    glVertex3f(1.0,2.85,0.03);
    glVertex3f(1.0,-3.301,0.03);
    glVertex3f(1.0,2.85,-0.03);
    glVertex3f(1.0,-3.301,-0.03);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.24,0.21,0.10);
    glNormal3f(0.0,0.0,1.0);
    glVertex3f(1.0,2.85,-0.03);
    glVertex3f(1.0,-3.301,-0.03);
    glVertex3f(0.97,2.85,-0.03);
    glVertex3f(0.97,-3.301,-0.03);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.26,0.23,0.12);
    glNormal3f(-1.0,0.0,0.0);
    glVertex3f(0.97,2.85,-0.03);
    glVertex3f(0.97,-3.301,-0.03);
    glVertex3f(0.97,2.85,-0.09);
    glVertex3f(0.97,-3.301,-0.09);
    glEnd();
    
    // Draw top part of doorframe
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.38,0.35,0.24);
    glNormal3f(0.0,1.0,0.0);
    glVertex3f(-1.091,2.85,0.0);
    glVertex3f(1.091,2.85,0.0);
    glVertex3f(-1.091,2.85,0.03);
    glVertex3f(1.091,2.85,0.03);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.30,0.27,0.16);
    glNormal3f(0.0,0.0,1.0);
    glVertex3f(-1.091,2.85,0.03);
    glVertex3f(1.091,2.85,0.03);
    glVertex3f(-1.091,2.76,0.03);
    glVertex3f(1.091,2.76,0.03);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.22,0.19,0.08);
    glNormal3f(0.0,-1.0,0.0);
    glVertex3f(-1.091,2.76,0.03);
    glVertex3f(1.091,2.76,0.03);
    glVertex3f(-1.091,2.76,-0.03);
    glVertex3f(1.091,2.76,-0.03);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.24,0.21,0.10);
    glNormal3f(0.0,0.0,1.0);
    glVertex3f(-1.091,2.76,-0.03);
    glVertex3f(1.091,2.76,-0.03);
    glVertex3f(-1.091,2.73,-0.03);
    glVertex3f(1.091,2.73,-0.03);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(0.22,0.19,0.08);
    glNormal3f(0.0,-1.0,0.0);
    glVertex3f(-1.091,2.73,-0.03);
    glVertex3f(1.091,2.73,-0.03);
    glVertex3f(-1.091,2.73,-0.09);
    glVertex3f(1.091,2.73,-0.09);
    glEnd();
    
    // Draw door
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,textures[TEX_WOOD]);
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(1.0,1.0,1.0);
    glNormal3f(0.0,0.0,1.0);
    glTexCoord2f(0.0,0.0);glVertex3f(-1.0,-3.301,-0.09);
    glTexCoord2f(0.0,1.0);glVertex3f(-1.0,2.76,-0.09);
    glTexCoord2f(1.0,0.0);glVertex3f(1.0,-3.301,-0.09);
    glTexCoord2f(1.0,1.0);glVertex3f(1.0,2.76,-0.09);
    glEnd();
}

void DrawGLScene(struct ViewState *vs, int *data)
{
	int i;

    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
    glRotatef(360.0-vs->xang,1.0,0.0,0.0);
    glRotatef(360.0-vs->yang,0.0,1.0,0.0);
    glTranslatef(-vs->xpos,-vs->ypos,-vs->zpos);
    
    glPushMatrix();
    glRotatef(-90.0,0.0,1.0,0.0);
    glTranslatef((float)NUM_HALL_SEGMENTS*(3.21+1.09)-3.21,0.0,-1.95);
    DrawWallSect(data[5],data[4]);
    glTranslatef(-3.21-1.09,0.0,0.0);
    DrawDoorSect(data[4]);

	for (i=1;i<NUM_HALL_SEGMENTS;i++) {
       glTranslatef(-1.09-3.21,0.0,0.0);
       DrawWallSect(data[i*6+5],data[i*6+4]);
       glTranslatef(-3.21-1.09,0.0,0.0);
       DrawDoorSect(data[i*6+4]);
    }
//    glTranslatef(-1.09-3.21,0.0,0.0);
//    DrawWallSect(0,0);
    glPopMatrix();

    glPushMatrix();
    glRotatef(90.0,0.0,1.0,0.0);
    glTranslatef((float)NUM_HALL_SEGMENTS*(-3.21-1.09)+3.21,0.0,-1.95);
    DrawWallSect(data[0],data[1]);
    glTranslatef(3.21+1.09,0.0,0.0);
    DrawDoorSect(data[1]);
    
    for (i=1;i<NUM_HALL_SEGMENTS;i++) {
       glTranslatef(1.09+3.21,0.0,0.0);
       DrawWallSect(data[i*6],data[i*6+1]);
       glTranslatef(3.21+1.09,0.0,0.0);
       DrawDoorSect(data[i*6+1]);
    }
//    glTranslatef(1.09+3.21,0.0,0.0);
//    DrawWallSect(0,0);
    glPopMatrix();        

	glPushMatrix();
	
	glTranslatef(0.0,0.0,(float)(NUM_HALL_SEGMENTS-1)*(1.09+3.21));
	DrawCeilingSect(data[2],data[3]);

	for(i=1;i<NUM_HALL_SEGMENTS;i++) {
		glTranslatef(0.0,0.0,-2.18-6.42);
		DrawCeilingSect(data[i*6+2],data[i*6+3]);
	}
	
	glPopMatrix();

    // Floor
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,textures[TEX_CARPET]);
    glBegin(GL_TRIANGLE_STRIP);
    glColor3f(1.0,1.0,1.0);
    glNormal3f(0.0,1.0,0.0);
    glTexCoord2f(0.0,0.0);glVertex3f(-2.042,-3.3,4.3*(float)NUM_HALL_SEGMENTS);
    glTexCoord2f(3.0,0.0);glVertex3f(2.042,-3.3,4.3*(float)NUM_HALL_SEGMENTS);
    glTexCoord2f(0.0,7.5*(float)NUM_HALL_SEGMENTS);
      glVertex3f(-2.042,-3.3,-4.3*(float)NUM_HALL_SEGMENTS);
    glTexCoord2f(3.0,7.5*(float)NUM_HALL_SEGMENTS);
      glVertex3f(2.042,-3.3,-4.3*(float)NUM_HALL_SEGMENTS);
    glEnd();

    SDL_GL_SwapBuffers();
}

int UpdateViewState(struct ViewState *vs)
{
    SDL_Event event;
    double time,new_vel_comp,ex,xtv,ytv,ztv,xang,yang;
    Uint32 ticks;
    static int xv=0,yv=0,zv=0,dirs=0;
    static Uint32 last_ticks=0;
    const double pyth_table[7]={0.0,                  // Zero directions (??)
                                1.0*VELOCITY,         // 1=sqrt(x^2)
                                0.70710678*VELOCITY,  // 1=sqrt(x^2+x^2)
                                0.57735027*VELOCITY,  // 1=sqrt(x^2+x^2+x^2)
                                0.5*VELOCITY,         // Shouldn't happen
                                0.44721360*VELOCITY,  // Shouldn't happen
                                0.40824829*VELOCITY}; // Shouldn't happen
    while(SDL_PollEvent(&event))
        switch(event.type)
        {
        case SDL_QUIT:
            return 1;
            break;
        case SDL_KEYDOWN:
            switch(event.key.keysym.sym)
            {
            case SDLK_ESCAPE:
                return 1;
                break;
            case SDLK_s:
                xv-=1;
                dirs++;
                break;
            case SDLK_d:
                xv+=1;
                dirs++;
                break;
            case SDLK_a:
                yv+=1;
                dirs++;
                break;
            case SDLK_z:
                yv-=1;
                dirs++;
                break;
            case SDLK_UP:
                zv-=1;
                dirs++;
                break;
            case SDLK_DOWN:
                zv+=1;
                dirs++;
                break;
            default:
                break;
            }
            break;
        case SDL_KEYUP:
            switch(event.key.keysym.sym)
            {
            case SDLK_s:
                xv+=1;
                dirs--;
                break;
            case SDLK_d:
                xv-=1;
                dirs--;
                break;
            case SDLK_a:
                yv-=1;
                dirs--;
                break;
            case SDLK_z:
                yv+=1;
                dirs--;
                break;
            case SDLK_UP:
                zv+=1;
                dirs--;
                break;
            case SDLK_DOWN:
                zv-=1;
                dirs--;
                break;
            default:
                break;
            }
            break;
#ifdef FULLSCREEN
        case SDL_MOUSEMOTION:
            vs->xang+=YSENS*(double)event.motion.yrel;
            vs->yang-=XSENS*(double)event.motion.xrel;
            if(vs->xang>90.0)
                vs->xang=90.0;
            else if(vs->xang<-90.0)
                vs->xang=-90.0;
            while(vs->yang>=360.0)
                vs->yang-=360.0;
            while(vs->yang<0.0)
                vs->yang+=360.0;
            break;
        case SDL_MOUSEBUTTONDOWN:
            switch(event.button.button)
            {
            case SDL_BUTTON_MIDDLE:
                zv-=1;
                dirs++;
                break;
            case SDL_BUTTON_RIGHT:
                zv+=1;
                dirs++;
                break;
            }
            break;
        case SDL_MOUSEBUTTONUP:
            switch(event.button.button)
            {
            case SDL_BUTTON_MIDDLE:
                zv+=1;
                dirs--;
                break;
            case SDL_BUTTON_RIGHT:
                zv-=1;
                dirs--;
                break;
            }
            break;
#endif
        default:
            break;
        }
    ticks=SDL_GetTicks();
    time=(double)(ticks-last_ticks)/1000.0;
    last_ticks=ticks;
    ex=exp(-time*FILTER_CONST);
    
    // Update velocity and position
    ytv=(double)yv;
    yang=vs->yang*PI_OVER_180;
    xang=vs->xang*PI_OVER_180;
    if(xv)
    {
        if(xv>0)
        {
            xtv=cos(yang);
            ztv=-sin(yang);
        }
        else
        {
            xtv=-cos(yang);
            ztv=sin(yang);
        }
    }
    else
        xtv=ztv=0.0;
    if(zv)
    {
        if(zv>0)
        {
            ytv-=sin(xang);
            new_vel_comp=cos(xang);
        }
        else
        {
            ytv+=sin(xang);
            new_vel_comp=-cos(xang);
        }
        ztv+=cos(yang)*new_vel_comp;
        xtv+=sin(yang)*new_vel_comp;
    }
    new_vel_comp=pyth_table[dirs]*(1.0-ex);
    xtv*=new_vel_comp;
    ytv*=new_vel_comp;
    ztv*=new_vel_comp;
    vs->xvel=ex*vs->xvel+xtv;
    vs->xpos+=vs->xvel*time;
    vs->yvel=ex*vs->yvel+ytv;
    vs->ypos+=vs->yvel*time;
    vs->zvel=ex*vs->zvel+ztv;
    vs->zpos+=vs->zvel*time;
    
    // Clip to within bounds
    if(vs->xpos>XCLIP)
        vs->xpos=XCLIP;
    else if(vs->xpos<-XCLIP)
        vs->xpos=-XCLIP;
    if(vs->ypos>YCLIP)
        vs->ypos=YCLIP;
    else if(vs->ypos<-YCLIP)
        vs->ypos=-YCLIP;
    if(vs->zpos>ZCLIP)
        vs->zpos=ZCLIP;
    else if(vs->zpos<-ZCLIP)
        vs->zpos=-ZCLIP;
//    printf("pos: <%f,%f,%f> vel: <%f,%f,%f> ang: <%f,%f>\r",vs->xpos,vs->ypos,vs->zpos,
//      vs->xvel,vs->yvel,vs->zvel,vs->yang,vs->xang);
    fflush(stdout);
    return 0;
}

static int light_data[6*NUM_HALL_SEGMENTS];
static struct ViewState viewstate;
static int start,frames=0;

int init_lights(void)
{
    SDL_Event dummy;
	int i;
    
    if(SDL_Init(SDL_INIT_VIDEO)<0)
    {
        fprintf(stderr,"Unable to init SDL: %s\n",SDL_GetError());
        return 1;
    }
    if(InitGL(WIDTH,HEIGHT))
        return 1;
#ifdef FULLSCREEN
    SDL_WM_GrabInput(SDL_GRAB_ON);
    SDL_ShowCursor(SDL_DISABLE);
#endif
    while(SDL_PollEvent(&dummy));   // I keep getting large mouse movement
                                    // events here (?)
    viewstate.xpos=viewstate.xang=viewstate.yang=
      viewstate.xvel=viewstate.yvel=viewstate.zvel=0.0;
    viewstate.ypos=1.5;
    viewstate.zpos=ZCLIP;
    
    for (i=0;i<(6*NUM_HALL_SEGMENTS);i++) light_data[i]=0;
    
    return 0;
}

void deinit_lights()
{
    SDL_Quit();
    printf("%f fps.\n",(double)frames/
      ((double)(SDL_GetTicks()-start)/1000.0));
}


int renderFrame(int *data)
{    

   DrawGLScene(&viewstate,data);
   if(UpdateViewState(&viewstate)) {
		deinit_lights();
		exit(0);
   }
   return 0;
}



void light_state(int light, int state)
{
	int i;
	static Uint32 ticks=0;
	for (i=0;i<6;i++) {
		light_data[light*6+i]=state*255;
	}
	
	if(SDL_GetTicks()>=ticks)
	{
	    renderFrame(light_data);
	    ticks+=FRAME_TIME;
	}
}
