#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include <SDL/SDL.h>

#include "interface_2001.h"

#define WIDTH 60
#define HEIGHT 100
#define PADDLE_WIDTH 15
#define PADDLE_HEIGHT 10
#define BALL_SIZE 10
#define FRAME_RATE 30
#define RESET_DELAY 2.0
#define START_SPEED 2.0
#define SPEED_STEP 0.1
#define MIN_Y_SPEED 0.5
#define SERVER_PORT 7880

unsigned char frmbuf[HEIGHT][WIDTH];
int master_light_data[60];

double ball_pos_x,ball_pos_y,ball_spd_x,ball_spd_y;
int pdl_pos_1,pdl_pos_2,last_pdl1_pos,last_pdl2_pos;
unsigned long frames,last_pdl1_frame,last_pdl2_frame;
SDL_Surface *screen;

void start_game(void)
{
    double ang;
    ball_pos_x=WIDTH>>1;
    ball_pos_y=HEIGHT>>1;
    do
    {
        ang=(double)rand()*2.0*3.141592653589/(double)RAND_MAX;
        ball_spd_y=START_SPEED*sin(ang);
    } while ((ball_spd_y>0?ball_spd_y:-ball_spd_y)<MIN_Y_SPEED);
    ball_spd_x=START_SPEED*cos(ang);
}

void do_physics(void)
{
    int t;
    if(ball_pos_x<0)
    {
//        printf("Waiting to start new game...\n");
        if(--ball_pos_x<=-RESET_DELAY*(double)FRAME_RATE)
        {
            start_game();
            printf("Starting new game...\n");
        }
        return;
    }
    ball_pos_x+=ball_spd_x;
    ball_pos_y+=ball_spd_y;
//    printf("Ball pos %f,%f; speed %f,%f\n",ball_pos_x,ball_pos_y,
//      ball_spd_x,ball_spd_y);
    t=BALL_SIZE>>1;
    if( ((ball_pos_x<=t)&&(ball_spd_x<0)) ||
        ((ball_pos_x>=WIDTH-t)&&(ball_spd_x>0)) )
    {
        printf("Ball hit a wall\n");
        if(ball_spd_x<0)
            ball_spd_x=-ball_spd_x+SPEED_STEP;
        else
            ball_spd_x=-ball_spd_x-SPEED_STEP;
    }
    t=PADDLE_HEIGHT+(BALL_SIZE>>1);
    if((ball_pos_y<=t)&&(ball_spd_y<0))
    {
        t=ball_pos_x-pdl_pos_1;
        if(t<0)
            t=-t;
        if(t<(PADDLE_WIDTH+BALL_SIZE)>>1)
        {
            ball_spd_y=-ball_spd_y+SPEED_STEP;
            ball_spd_x=((ball_spd_x*3.0)+(double)(pdl_pos_1-last_pdl1_pos)/
                        (double)(frames-last_pdl1_frame))/4.0;
            printf("Ball hit paddle 1\n");
        }
    }
    else if((ball_pos_y>=HEIGHT-t)&&(ball_spd_y>0))
    {
        t=ball_pos_x-pdl_pos_2;
        if(t<0)
            t=-t;
        if(t<(PADDLE_WIDTH+BALL_SIZE)>>1)
        {
            ball_spd_y=-ball_spd_y-SPEED_STEP;
            ball_spd_x=((ball_spd_x*3.0)+(double)(pdl_pos_2-last_pdl2_pos)/
                        (double)(frames-last_pdl2_frame))/4.0;
            printf("Ball hit paddle 2\n");
        }
    }
    if((ball_pos_y<=0)||(ball_pos_y>=HEIGHT))
    {
        ball_pos_x=-1;
        ball_pos_y=-BALL_SIZE;
        printf("Ball lost\n");
    }
}

void render_frame(void)
{
    int y,x,ball,paddle,t;
    if(ball_pos_x<0)
    {
        if((int)ball_pos_x&16)
            for(x=0;x<60;x++)
                master_light_data[x]=255;
        else
            for(x=0;x<60;x++)
                master_light_data[x]=0;
        renderFrame(master_light_data);
        return;
    }
    memset(master_light_data,0,60*sizeof(int));
    for(y=0;y<HEIGHT;y++)
    {
        t=y-ball_pos_y;
        if(t<0)
            t=-t;
        if(t<(BALL_SIZE>>1))
            ball=1;
        else
            ball=0;
        if(y<=PADDLE_HEIGHT)
            paddle=1;
        else if(y>=(HEIGHT-PADDLE_HEIGHT))
            paddle=2;
        else
            paddle=0;
        for(x=0;x<WIDTH;x++)
        {
            if((y>PADDLE_HEIGHT)&&(y<(HEIGHT-PADDLE_HEIGHT)))
                frmbuf[y][x]=(frmbuf[y][x]*7)>>3;
            else
                frmbuf[y][x]=0;
            if(ball)
            {
                t=x-ball_pos_x;
                if(t<0)
                    t=-t;
                if(t<(BALL_SIZE>>1))
                    frmbuf[y][x]=255;
            }
            if(paddle)
            {
                if(paddle==2)
                    t=x-pdl_pos_2;
                else
                    t=x-pdl_pos_1;
                if(t<0)
                    t=-t;
                if(t<(PADDLE_WIDTH>>1))
                    frmbuf[y][x]=255;
            }
            master_light_data[(y/10)*6+x/10]+=frmbuf[y][x];
            ((unsigned char *)screen->pixels)[y*screen->pitch+x]=frmbuf[y][x];
        }
    }
    for(y=0;y<60;y++)
        master_light_data[y]/=100;
    renderFrame(master_light_data);
    SDL_UpdateRect(screen,0,0,0,0);
}

int open_udp_server(int port)
{
        int sock;
        struct sockaddr_in my_addr;

        if ((sock=socket(AF_INET, SOCK_DGRAM, 0))==-1) {
                perror("start_net: creating socket");
                return -1;
        }

        my_addr.sin_family=AF_INET;
        my_addr.sin_addr.s_addr=INADDR_ANY;
        my_addr.sin_port=htons(port);
        bzero(&(my_addr.sin_zero),8);

        if (bind(sock, (struct sockaddr *)&my_addr, sizeof(struct sockaddr))
<0)
 {
                perror("start_net: binding socket");
                fprintf(stderr,"ERROR!\n");
                close(sock);
                return -1;
        }

        return sock;
}

int main(void)
{
    int x,y,sock1,sock2;
    Uint32 ticks,frame_time;
    fd_set fsr;
    char buf[5];
    struct timeval tv;
    SDL_Color colors[256];
    server_port=SERVER_PORT;
    if(init_lights())
    {
        fprintf(stderr,"init_lights failed!\n");
        return -1;
    }
    srand(time(NULL));
    if(SDL_Init(SDL_INIT_VIDEO)<0)
    {
        fprintf(stderr,"Unable to init SDL: %s\n",SDL_GetError());
        return 1;
    }
    if((screen=SDL_SetVideoMode(WIDTH,HEIGHT,8,0))==NULL)
    {
        fprintf(stderr,"Unable to create window: %s\n",SDL_GetError());
        SDL_Quit();
        return 2;
    }
    atexit(SDL_Quit);
    for(x=0;x<256;x++)
        colors[x].r=colors[x].g=colors[x].b=x;
    SDL_SetPalette(screen,SDL_LOGPAL,colors,0,256);
    for(y=0;y<HEIGHT;y++)
        for(x=0;x<WIDTH;x++)
            frmbuf[y][x]=0;
    pdl_pos_1=pdl_pos_2=last_pdl1_pos=last_pdl2_pos=WIDTH>>1;
    start_game();
    last_pdl1_frame=last_pdl2_frame=0;
    if((sock1=open_udp_server(8000))<0)
    {
        perror("open_udp_server(8000)");
        exit(1);
    }
    if((sock2=open_udp_server(8001))<0)
    {
        perror("open_udp_server(8001)");
        exit(1);
    }
    frame_time=1000/FRAME_RATE;
    ticks=SDL_GetTicks()+frame_time;
    for(frames=0;;frames++)
    {
        do_physics();
        render_frame();
        while(SDL_GetTicks()<ticks)
        {
            FD_ZERO(&fsr);
            FD_SET(sock1,&fsr);
            FD_SET(sock2,&fsr);
            tv.tv_sec=0;
            tv.tv_usec=10000;
            select(FD_SETSIZE,&fsr,NULL,NULL,&tv);
            if(FD_ISSET(sock1,&fsr))
            {
                read(sock1,buf,5);
                last_pdl1_pos=pdl_pos_1;
                pdl_pos_1=WIDTH-((((*(long *)buf)*(WIDTH>>1))>>15)+(WIDTH>>1));
                last_pdl1_frame=frames;
            }
            if(FD_ISSET(sock2,&fsr))
            {
                read(sock2,buf,5);
                last_pdl2_pos=pdl_pos_2;
                pdl_pos_2=(((*(long *)buf)*(WIDTH>>1))>>15)+(WIDTH>>1);
                last_pdl2_frame=frames;
            }
        }
        ticks+=frame_time;
    }
    // Not going to get here just yet
    deinit_lights();
}
