#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "interface_2001.h"

int master_light_data[60];

int count_usecs(struct timeval *tv);
static int paddle_1_port=8000;
static int paddle_2_port=8001;

static int paddle_1_sock, paddle_2_sock;


int main()
{
	int ret;
	
	server_port=7880;
	if (init_lights()) {
		fprintf(stderr,"init_lights failed!\n");
		return -1;
	}

	srand(time(NULL));
	
	auto_pong_init(NULL);

	for (;;) {
		auto_pong_frame(master_light_data);
		renderFrame(master_light_data);
		usleep(33000);
	}
	
	auto_pong_deinit();
	deinit_lights();
}


static float pong_paddle_1_pos=0;
static float pong_paddle_2_pos=0;
static float pong_ball_x_pos=3;
static float pong_ball_y_pos=5;
static float pong_ball_x_speed=.1;
static float pong_ball_y_speed=.1;
static int pong_counter;
static int pong_ball_counter;
static int pong_loss_count;


int auto_pong_init(char *param)
{

	pong_counter=0;
	pong_ball_counter=60;	// this is used for inter_game pauses
	pong_loss_count=0;
	pong_ball_x_pos=rand()%6;
	pong_ball_y_pos=5;
	return 0;
}

int auto_pong_deinit(void)
{

	return 0;
}

int auto_pong_frame(int light_data[NUM_SEGMENTS*6])
{
	int i,x,y;
	float ft, fraction;

	pong_counter++;
	if (pong_ball_counter>0) pong_ball_counter--;

	if (pong_ball_counter==0) {

    	// move the ball, check for bounces or end of game...
    	// then move the paddles
    	
    	pong_ball_x_pos+=pong_ball_x_speed;
    	if (pong_ball_x_pos>5) {
    		pong_ball_x_pos=5;
    		pong_ball_x_speed=-pong_ball_x_speed;
    	} else if (pong_ball_x_pos<0) {
    		pong_ball_x_pos=0;
    		pong_ball_x_speed=-pong_ball_x_speed;
    	}
    	
    	
    	pong_ball_y_pos+=pong_ball_y_speed;
    	if (pong_ball_y_pos>9) {  // we are in the last row
    		x=pong_ball_x_pos;
    		if (x==pong_paddle_2_pos) { // we hit the other paddle
    			pong_ball_y_pos=9;
    			pong_ball_y_speed=-pong_ball_y_speed;
    		} else { // we missed the paddle
    			pong_ball_y_pos=5;
    			pong_ball_y_speed=-pong_ball_y_speed;
    			pong_ball_counter=60;
    			pong_ball_x_pos=rand()%6;
    			pong_loss_count++;
    		}
    	}
    	if (pong_ball_y_pos<1) { // we are in the first row
    		x=pong_ball_x_pos;
    		if (pong_ball_x_pos>(pong_paddle_1_pos-1) && pong_ball_x_pos<(pong_paddle_1_pos+1)) { // we hit the paddle
    			pong_ball_y_pos=1;
    			pong_ball_y_speed=-pong_ball_y_speed;
//    			if (abs(pong_ball_x_pos-pong_paddle_1_pos)>.5) pong_ball_x_speed*=1.2;
//    			if (abs(pong_ball_x_pos-pong_paddle_1_pos)<.3) pong_ball_y_speed*=1.2;
    		} else {
    			pong_ball_y_pos=5;
    			pong_ball_x_pos=rand()%6;
    			pong_ball_y_speed=-pong_ball_y_speed;
    			pong_ball_counter=60;
    			pong_loss_count++;
    		}
    	}
	}

	if (pong_ball_y_pos<5 && pong_ball_y_speed<0) {
    	x=rand()%5+2;
    	if ((pong_counter%x)==0) {
    		x=pong_ball_x_pos;
    		if ((x-pong_paddle_1_pos)<0) {
    			pong_paddle_1_pos-=.33;
    		} else if ((x-pong_paddle_1_pos)>0) {
    			pong_paddle_1_pos+=.33;
    		}
    	}
	}	
	
	if (pong_ball_y_pos>5 && pong_ball_y_speed>0) {
    	x=rand()%2+1;
    	if ((pong_counter%x)==0) {
    		x=pong_ball_x_pos;
    		if ((x-pong_paddle_2_pos)<0) {
    			pong_paddle_2_pos--;
    		} else if ((x-pong_paddle_2_pos)>0) {
    			pong_paddle_2_pos++;
    		}
    	}
    }


	
	// render the new scene

	memset(light_data,0,sizeof(int)*NUM_SEGMENTS*6);
	
	ft=175+75.0*sin((float)pong_counter/5.0);
	x=pong_paddle_1_pos;
	fraction=pong_paddle_1_pos-(float)x;
	
	light_data[x]=ft*(1-fraction);
	light_data[x+1]=ft*fraction;
	
	light_data[(NUM_SEGMENTS-1)*6+(int)pong_paddle_2_pos]=175+75.0*sin((float)(pong_counter+2)/5.0);

	if (pong_ball_counter==0) {
	
		x=pong_ball_x_pos;
		y=pong_ball_y_pos;
		
		light_data[y*6+x]=255;
	}

	if (pong_loss_count>2) return -1;

	return 0;
}
		