/*  XMMS - Cross-platform multimedia player
 *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "config.h"

#include <gtk/gtk.h>
#include <math.h>
#include "../../interface_2001.h"
#include "pattern.h"

#include <xmms/plugin.h>
#include "xmms_logo.xpm"

#include <time.h>
#include <sys/time.h>
#include <stdio.h>

#define NUM_BANDS 10

static GtkWidget *window = NULL,*area;
static GdkPixmap *bg_pixmap = NULL, *draw_pixmap = NULL, *bar = NULL;
static GdkGC *gc = NULL;
static gint16 bar_heights[NUM_BANDS];
static gdouble bar_rate[NUM_BANDS];
static gdouble bar_averages[NUM_BANDS];
static gdouble bar_scales[NUM_BANDS];
static gint timeout_tag;
static gdouble scale;
static gint light_data[NUM_SEGMENTS*6];

static struct pattern_info patterns[] = { 
//		 { simple_equalizer_init, simple_equalizer_frame, simple_equalizer_deinit },
		 { sine_wave_init, sine_wave_frame, sine_wave_deinit },
//		 { ring_equalizer_init, ring_equalizer_frame, ring_equalizer_deinit },
//		 { trailing_lines_init, trailing_lines_frame, trailing_lines_deinit },
		 { rise_init, rise_frame, rise_deinit },
//		 { bar_equalizer_init, bar_equalizer_frame, bar_equalizer_deinit },
		 { random_spots_init, random_spots_frame, random_spots_deinit },
//		 { auto_pong_init, auto_pong_frame, auto_pong_deinit },
//		 { center_meter_init, center_meter_frame, center_meter_deinit },
		 { brightness_level_init, brightness_level_frame, brightness_level_deinit },
		 { brightness_level_init, brightness_level_frame, brightness_level_deinit },
////		 { vertical_meter_init, vertical_meter_frame, vertical_meter_deinit },
//		 { swirl_meter_init, swirl_meter_frame, swirl_meter_deinit },
//		 { vertical_spectrum_init, vertical_spectrum_frame, vertical_spectrum_deinit },
		 { spinning_stripe_init, spinning_stripe_frame, spinning_stripe_deinit },
//		 { warp_core_init, warp_core_frame, warp_core_deinit },
//		 { explosion_init, explosion_frame, explosion_deinit },
		 {NULL, NULL, NULL}
		};
static gint num_patterns;

#define NUM_BEATS 20
static struct timeval beat_ts[NUM_BEATS];
static int beat_ts_index=0;
static float beats_pm=0;

static void anim_init(void);
static void anim_cleanup(void);
static void anim_playback_start(void);
static void anim_playback_stop(void);
static void anim_analyze_freq(gint16 data[2][256]);
static void anim_render_freq(gint16 data[2][256]);

VisPlugin anim_vp =
{
	NULL,
	NULL,
	0,
	NULL, /* Description */
	0,
	1,		
	anim_init, /* init */
	anim_cleanup, /* cleanup */
	NULL, /* about */
	NULL, /* configure */
	NULL, /* disable_plugin */
	anim_playback_start, /* playback_start */
	anim_playback_stop, /* playback_stop */
	NULL, /* render_pcm */
	anim_render_freq  /* render_freq */
};

VisPlugin *get_vplugin_info(void)
{
	anim_vp.description =
		g_strdup_printf("X-Mas Lights Controller %s", VERSION);
	return &anim_vp;
}

#define WIDTH 250
#define HEIGHT 100


static void anim_destroy_cb(GtkWidget *w,gpointer data)
{
	anim_vp.disable_plugin(&anim_vp);
}

static void anim_init(void)
{
	GdkColor color;
	int i;
	if(window)
		return;
	window = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_title(GTK_WINDOW(window),"Spectrum analyzer");
	gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, FALSE);
	gtk_widget_realize(window);
	bg_pixmap = gdk_pixmap_create_from_xpm_d(window->window,NULL,NULL,sanalyzer_xmms_logo_xpm);
	gdk_window_set_back_pixmap(window->window,bg_pixmap,0);
	gtk_signal_connect(GTK_OBJECT(window),"destroy",GTK_SIGNAL_FUNC(anim_destroy_cb),NULL);
	gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &window);
	gtk_widget_set_usize(window, WIDTH, HEIGHT);
	gc = gdk_gc_new(window->window);
	draw_pixmap = gdk_pixmap_new(window->window,WIDTH,HEIGHT,gdk_rgb_get_visual()->depth);
	
	bar = gdk_pixmap_new(window->window,25, HEIGHT, gdk_rgb_get_visual()->depth);
	for(i = 0; i < HEIGHT / 2; i++)
	{
		color.red = 0xFFFF;
		color.green = ((i * 255) / (HEIGHT / 2)) << 8;
		color.blue = 0;
		
		gdk_color_alloc(gdk_colormap_get_system(),&color);
		gdk_gc_set_foreground(gc,&color);
		gdk_draw_line(bar,gc,0,i,24,i);
	}
	for(i = 0; i < HEIGHT / 2; i++)
	{
		color.red = (255 - ((i * 255) / (HEIGHT / 2))) <<8;
		color.green = 0xFFFF;
		color.blue = 0;
		
		gdk_color_alloc(gdk_colormap_get_system(),&color);
		gdk_gc_set_foreground(gc,&color);
		gdk_draw_line(bar,gc,0,i + (HEIGHT / 2),24,i + (HEIGHT / 2));
	}
	scale = HEIGHT / log(256);
	gdk_color_black(gdk_colormap_get_system(),&color);
	gdk_gc_set_foreground(gc,&color);
	
	area = gtk_drawing_area_new();
	gtk_container_add(GTK_CONTAINER(window),area);
	gtk_widget_realize(area);
	gdk_window_set_back_pixmap(area->window,bg_pixmap,0);
	
	gtk_widget_show(area);
	gtk_widget_show(window);
	gdk_window_clear(window->window);
	gdk_window_clear(area->window);

	for (i=0;i<NUM_BANDS;i++) {
		bar_averages[i]=64.0;
		bar_scales[i]=1.0;
	}

	num_patterns=0;
	while (patterns[num_patterns].init!=NULL) num_patterns++;

	init_lights();
	srand(time(NULL));
	
	memset(&beat_ts,0,sizeof(struct timeval)*NUM_BEATS);
}

static void anim_cleanup(void)
{
 	if(window)
	{
		gtk_widget_destroy(window);
	}
	if(gc)
	{
		gdk_gc_unref(gc);
		gc = NULL;
	}
	if(bg_pixmap)
	{
		gdk_pixmap_unref(bg_pixmap);
		bg_pixmap = NULL;
	}
	if(draw_pixmap)
	{
		gdk_pixmap_unref(draw_pixmap);
		draw_pixmap = NULL;
	}
	if(bar)
	{
		gdk_pixmap_unref(bar);
		bar = NULL;
	}

	deinit_lights();
}

static gint draw_func(gpointer data)
{
	gint i,i2;
	
	if(!window)
	{
		timeout_tag = 0;
		return FALSE;
	}
	
	GDK_THREADS_ENTER();

	gdk_draw_rectangle(draw_pixmap,gc,TRUE,0,0,WIDTH,HEIGHT);

	
	for(i = 0; i < NUM_BANDS; i++)
	{
		gdk_draw_pixmap(draw_pixmap,gc,bar, 0,HEIGHT - 1 - bar_heights[i], i * (WIDTH / NUM_BANDS), HEIGHT - 1 - bar_heights[i], (WIDTH / NUM_BANDS) - 1, bar_heights[i]);
		
	}
	gdk_window_clear(area->window);


	GDK_THREADS_LEAVE();
	
	return TRUE;
}

static void anim_playback_start(void)
{
	if(window)
	{
		gdk_window_set_back_pixmap(area->window,draw_pixmap,0);
		gdk_window_clear(area->window);
	}

}


static void anim_playback_stop(void)
{
	if(GTK_WIDGET_REALIZED(area))
	{
		gdk_window_set_back_pixmap(area->window,bg_pixmap,0);
		gdk_window_clear(area->window);
	}
}


static void anim_render_freq(gint16 data[2][256])
{
	static gint current_pattern=-1;
	static gint time_left=10;
	static gint frame_time=0;
	int ret;


	// first analyze the data so that all patterns have something to work with
	anim_analyze_freq(data);
	
	// when either the timer runs out, or the pattern says it is done we need
	// to select a new pattern.  This should have two modes. One a random pattern
	// is selected based on the previous one.  Two a script file is used to select the
	// next pattern.
	time_left--;
	if (time_left<0 && current_pattern>=0) {
		printf("calling deinit for pattern %i\n",current_pattern);
		(patterns[current_pattern].deinit)();
	}
	if (current_pattern==-1 || time_left<0) {
		// just select a random pattern for now
		ret=1;
		while (ret!=0) {
			current_pattern=rand()%num_patterns;
			printf("calling init for pattern %i\n",current_pattern);
			ret=(patterns[current_pattern].init)(NULL);
			printf("\n");
		}	
		time_left=125+rand()%250;
	}

	// now we need to run the correct pattern
	GDK_THREADS_ENTER();
	
	ret=(patterns[current_pattern].frame)(bar_averages,bar_heights,light_data,data, time_left);

	if (ret<0) {
		(patterns[current_pattern].deinit)();
		current_pattern=-1;
	}

	if (renderFrame(light_data)<0) {
		anim_vp.disable_plugin(&anim_vp);
	}
	
	GDK_THREADS_LEAVE();	

	draw_func(NULL);
	
}
	

static void anim_analyze_freq(gint16 data[2][256])
{
	gint i,c;
	gint y;
	long int ti;
	static gint flop=0;
	gdouble t;
	int old_bass;
	struct timeval tv,tv2;
	
	gint xscale[] = {0, 1, 2, 3, 5, 7, 10, 14, 20, 28, 40, 54, 74, 101, 137, 187, 255};
	
	
	if(!window)
		return;

	old_bass=bar_heights[0];
		
	for(i = 0; i < NUM_BANDS; i++)
	{
		for(c = xscale[i], y = 0; c < xscale[i + 1]; c++)
		{
			if(data[0][c] > y)
				y = data[0][c];
		}
		y >>= 7;
		
		t=0;
		if (y != 0) {
			t=log(y)*scale;
		}
			
		bar_averages[i]=(bar_averages[i]*15.0f+t)/16.0f;
		
		if(y != 0)
		{
			y = (gint)(log(y) * scale);
			if(y > HEIGHT - 1)
				y = HEIGHT - 1;
		}
		
		bar_scales[i]=min(bar_scales[i]+.002,3);
		
		if (i<10) {
			y-=((gint)bar_averages[i]);
			if (((double)y*bar_scales[i])>(HEIGHT)) 
				bar_scales[i]=(double)(HEIGHT)/(double)y;
			y*=bar_scales[i];
			y+=10;

			if (y<0) y=0;
			if (y> HEIGHT - 1) y= HEIGHT-1;
		}
 
		if(y > bar_heights[i]) {
			bar_heights[i] = y;
			bar_rate[i]=0;
		}
		else if(bar_heights[i] > bar_rate[i]) {
			bar_heights[i] -= bar_rate[i];
			bar_rate[i]+=0.25;
		} else
			bar_heights[i] = 0;
//		bar_heights[i]=(gint)(bar_averages[i]*(double)(HEIGHT/256.0));
		
	}
	
	for (i=0,c=0;i<NUM_BANDS;i++) if (bar_heights[i]<20) c++;
	
	if (c==NUM_BANDS) memset(&beat_ts,0,sizeof(struct timeval)*NUM_BEATS);
	
	if ((bar_heights[0]-old_bass)>20 && bar_heights[0]>30) {
		// detected a beat... insert timestamp into circular queue...
		gettimeofday(&beat_ts[beat_ts_index],NULL);
		beat_ts_index=(beat_ts_index+1)%NUM_BEATS;
	}
	
	tv.tv_sec=2000000000;
	tv.tv_usec=0;
	c=0;
	for (i=0;i<NUM_BEATS;i++) {
		if ((beat_ts[i].tv_sec>0)) {
			c++;
			if (timercmp(&beat_ts[i],&tv,<) ) {
				memcpy(&tv,&beat_ts[i],sizeof(struct timeval));
			}
		}
	}
	
	gettimeofday(&tv2,NULL);
	
	
	ti=((tv2.tv_sec-tv.tv_sec)*1000+(tv2.tv_usec-tv.tv_usec)/1000);

	if (ti>0) {
		beats_pm=(float)c/(float) ti ;
	} else {
		beats_pm=0;
	}
	beats_pm=beats_pm*60000.0f;

//	printf("beats per minute: %f\r",beats_pm);
//	fflush(stdout);
	
	draw_func(NULL);
	return;			
}
